from typing import Callable, List, Optional

import log
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys

from .compat import Page, PlaywrightError
from .types import GenericBrowser, PlaywrightBrowser

__all__ = ["browser", "client", "linebreak"]

browser: GenericBrowser = None
linebreak: bool = True


class _Client:
    @property
    def windows(self):
        if isinstance(browser, PlaywrightBrowser):
            return []

        return browser.windows

    @property
    def page(self) -> Page:
        # Raises an AttributeError for non-Playwright browsers
        return browser.contexts[0].pages[0]

    @property
    def url(self) -> str:
        if browser is None:
            return ""

        if isinstance(browser, PlaywrightBrowser):
            try:
                self.page.bring_to_front()
            except IndexError:
                return ""
            return self.page.url

        return browser.url

    @property
    def title(self) -> str:
        if isinstance(browser, PlaywrightBrowser):
            return self.page.title()

        return browser.title

    @property
    def html(self) -> str:
        if isinstance(browser, PlaywrightBrowser):
            return self.page.content()

        return browser.html

    @staticmethod
    def visit(url: str, size: Optional[dict]) -> None:
        exception = RuntimeError(f"Unable to load {url}")
        if isinstance(browser, PlaywrightBrowser):
            page = browser.new_page(screen=size, viewport=size)  # type: ignore
            try:
                page.goto(url)
            except PlaywrightError:
                raise exception from None
        else:
            if size and browser.driver.get_window_size() != size:
                browser.driver.set_window_size(size["width"], size["height"])
                browser.driver.set_window_position(0, 0)
                size = browser.driver.get_window_size()
                log.debug(f"Resized browser: {size}")
            try:
                browser.visit(url)
            except WebDriverException as e:
                log.error(e)
                raise exception from None

    def type_key(self, name: str) -> Callable:
        if isinstance(browser, PlaywrightBrowser):
            key = name.capitalize()
            return lambda: self.page.keyboard.press(key)

        key = getattr(Keys, name.upper())
        return ActionChains(browser.driver).send_keys(key).perform

    def type_key_with_modifier(self, names: List[str]) -> Callable:
        if len(names) > 2:
            raise ValueError("Multiple modifier keys are not yet supported")

        if isinstance(browser, PlaywrightBrowser):
            keys = "+".join(name.capitalize() for name in names)
            return lambda: self.page.keyboard.press(keys)

        modifier = getattr(Keys, names[0].upper())
        key = getattr(Keys, names[1].upper())
        return (
            ActionChains(browser.driver)
            .key_down(modifier)
            .send_keys(key)
            .key_up(modifier)
            .perform
        )

    def execute(self, javascript: str):
        if isinstance(browser, PlaywrightBrowser):
            self.page.evaluate(javascript)
        else:
            browser.execute_script(javascript)

    @staticmethod
    def clear_cookies():
        log.info("Clearing cookies")
        if isinstance(browser, PlaywrightBrowser):
            browser.contexts[0].clear_cookies()
        else:
            browser.cookies.delete_all()


client = _Client()



from urllib.parse import unquote

import log
from flask import redirect, request, url_for
from flask_api import FlaskAPI

from .api import visit

app = FlaskAPI("Pomace")


@app.route("/")
def index():
    return redirect("/sites?url=http://example.com")


@app.route("/sites")
def pomace():
    if "url" not in request.args:
        return redirect("/")

    url = request.args["url"]
    page = visit(url)

    transitioned = False
    for action, value in request.args.items():
        if "_" in action:
            page, transitioned = page.perform(action, value or "<missing>")
            if transitioned:
                log.info(f"Transitioned to {page}")

    link = unquote(url_for(".pomace", url=page.url, _external=True))
    if transitioned:
        return redirect(link)

    data = {
        "id": page.identity,
        "url": page.url,
        "title": page.title,
        "html": page.soup.prettify(),
        "text": page.text,
        "_self": link,
        "_actions": {action: link + "&" + action for action in dir(page)},
    }

    return data



import os
from importlib import reload

import log
from cleo.application import Application
from cleo.commands.command import Command, Verbosity
from cleo.helpers import argument, option
from startfile import startfile

from . import __version__, models, prompts, server, utils
from .config import settings

VERBOSITY = {
    Verbosity.QUIET: 0,
    Verbosity.NORMAL: 1,
    Verbosity.VERBOSE: 2,
    Verbosity.VERY_VERBOSE: 3,
    Verbosity.DEBUG: 4,
}


class BaseCommand(Command):
    arguments = [
        argument(
            "domain",
            description="Name of sites directory for this repository.",
            optional=True,
        ),
    ]
    options = [
        option(
            "framework",
            "f",
            description="Framework to control browsers.",
            flag=False,
        ),
        option(
            "browser",
            "b",
            description="Browser to use for automation.",
            flag=False,
        ),
        option(
            "headless",
            "d",
            description="Run the specified browser in a headless mode.",
        ),
        option(
            "prompt",
            "p",
            description="Prompt for secrets before running.",
            flag=False,
            multiple=True,
        ),
        option(
            "root",
            "r",
            description="Path to directory to containing models.",
        ),
    ]

    def handle(self):
        self.configure_logging()
        self.set_directory()
        utils.locate_models()
        self.update_settings()
        utils.launch_browser()
        try:
            self.handle_command()
        except KeyboardInterrupt:
            log.debug("User cancelled loop")
        finally:
            utils.close_browser()
            prompts.linebreak()

    def handle_command(self):
        raise NotImplementedError

    def configure_logging(self):
        log.reset()
        verbosity = VERBOSITY[self.io.output.verbosity]
        extra = 1 if self.name == "exec" else 0
        log.init(verbosity=verbosity + extra)
        log.silence("datafiles", allow_warning=True)
        log.silence("urllib3.connectionpool", allow_error=True)

    def set_directory(self):
        if self.option("root"):
            os.chdir(self.option("root"))

    def update_settings(self):
        if self.option("framework"):
            settings.framework = self.option("framework").lower()
        prompts.framework_if_unset()

        if self.option("browser"):
            settings.browser.name = self.option("browser").lower()
        prompts.browser_if_unset()

        settings.browser.headless = self.option("headless")

        for name in self.option("prompt"):
            prompts.secret_if_unset(name)

        if self.name == "exec":
            return

        if self.argument("domain"):
            if "://" in self.argument("domain"):
                settings.url = self.argument("domain")
            elif "." in self.argument("domain"):
                settings.url = "http://" + self.argument("domain")
            else:
                settings.url = ""
        domains = sorted(set(p.domain for p in models.Page.objects.all()))
        prompts.url_if_unset(domains)


class AliasCommand(BaseCommand):
    name = "alias"
    description = "Map one domain's site definitions to another."
    arguments = [
        argument(
            "source",
            description="Domain to map to another.",
        ),
        argument(
            "target",
            description="Domain with existing models.",
        ),
    ]
    options = []  # type: ignore

    def handle(self):
        self.configure_logging()
        self.handle_command()

    def handle_command(self):
        source = self.argument("source")
        target = self.argument("target")
        settings.aliases[source] = target
        self.line(
            f"<fg=white;options=bold>{source}</> mapped to "
            f"<fg=white;options=bold>{target}</>"
        )


class CleanCommand(BaseCommand):
    name = "clean"
    description = "Remove unused actions in local site definitions."
    options = [
        option(
            "root",
            "r",
            description="Path to directory to containing models.",
        )
    ]

    def handle(self):
        self.configure_logging()
        self.set_directory()
        utils.locate_models()
        self.handle_command()

    def handle_command(self):
        if self.argument("domain"):
            pages = models.Page.objects.filter(domain=self.argument("domain"))
        else:
            pages = models.Page.objects.all()
        for page in pages:
            page.clean(force=True)


class CloneCommand(BaseCommand):
    name = "clone"
    description = "Download site definitions from a Git repository."
    arguments = [
        argument(
            "url",
            description="Git repository URL containing pomace models.",
        ),
        argument(
            "domain",
            description="Name of sites directory for this repository.",
            optional=True,
        ),
    ]
    options = [
        option(
            "force",
            "f",
            description="Overwrite uncommitted changes in cloned repositories.",
        ),
        option(
            "root",
            "r",
            description="Path to directory to containing models.",
        ),
    ]

    def handle(self):
        self.configure_logging()
        self.set_directory()
        utils.locate_models()
        self.handle_command()

    def handle_command(self):
        utils.clone_models(
            self.argument("url"),
            domain=self.argument("domain"),
            force=self.option("force"),
        )


class EditCommand(BaseCommand):
    name = "edit"
    description = "Open the configuration file for editing."
    arguments = []  # type: ignore
    options = []  # type: ignore

    def handle(self):
        self.handle_command()

    def handle_command(self):
        startfile(settings.datafile.path)


class ExecCommand(BaseCommand):
    name = "exec"
    description = "Run a Python script with 'pomace' dependency."
    arguments = [
        argument(
            "script",
            description="Path to a Python script.",
        ),
    ]

    def handle_command(self):
        utils.run_script(self.argument("script"))


class RunCommand(BaseCommand):
    name = "run"
    description = "Start the command-line interface."

    def handle_command(self):
        self.clear_screen()
        page = models.auto()
        self.display_url(page)

        while True:
            action = prompts.action(page)
            if action:
                page, transitioned = page.perform(action)
                if transitioned:
                    self.clear_screen()
                    self.display_url(page)
            else:
                reload(models)
                self.clear_screen()
                page = models.auto()
                self.display_url(page)

    def clear_screen(self):
        os.system("cls" if os.name == "nt" else "clear")

    def display_url(self, page):
        self.line(f"<fg=white;options=bold>{page}</>")
        prompts.linebreak(force=True)


class ShellCommand(BaseCommand):
    name = "shell"
    description = "Launch an interactive shell."

    def handle_command(self):
        prompts.shell()


class ServeCommand(BaseCommand):
    name = "serve"
    description = "Start the web API interface."
    options = BaseCommand.options + [
        option(
            "debug",
            description="Run the server in debug mode.",
        ),
    ]

    def handle(self):
        self.configure_logging()
        self.set_directory()
        self.update_settings()
        prompts.bullet = None
        utils.locate_models()
        self.handle_command()

    def handle_command(self):
        try:
            server.app.run(debug=self.option("debug"))
        finally:
            utils.close_browser()


application = Application("pomace", __version__)
application.add(AliasCommand())
application.add(CleanCommand())
application.add(CloneCommand())
application.add(EditCommand())
application.add(ExecCommand())
application.add(RunCommand())
application.add(ServeCommand())
application.add(ShellCommand())



# pylint: disable=expression-not-assigned,unused-variable,redefined-outer-name,unused-argument

import pytest
import requests

from ..models import Action, Locator, Page


@pytest.fixture
def locator():
    return Locator("name", "email")


@pytest.fixture
def action():
    return Action("fill", "email")


@pytest.fixture
def page(mockbrowser):
    return Page("example.com", "/foo/bar")


def describe_locator():
    def describe_bool():
        def it_is_false_when_placeholder(expect, locator):
            expect(bool(locator)) == True
            locator.value = ""
            expect(bool(locator)) == False

    def describe_sort():
        def it_orders_by_uses(expect):
            locators = [
                Locator("name", "bbb", uses=5),
                Locator("name", "aaa", uses=6),
                Locator("name", "BBB", uses=6),
                Locator("name", "AAA", uses=7),
                Locator("name", "zzz", uses=8),
            ]
            expect(sorted(locators)) == locators  # type: ignore

    def describe_find():
        def it_returns_callable(expect, mockbrowser, locator):
            expect(locator.find()) == "mockelement:name=email"

        def it_can_find_links_by_partial_text(expect, mockbrowser, locator):
            locator.mode = "text (partial)"
            expect(locator.find()) == "mockelement:links.partial_text=email"

        def it_can_find_links_by_aria_label(expect, mockbrowser, locator):
            locator.mode = "aria-label"
            expect(locator.find()) == 'mockelement:css=[aria-label="email"]'

    def describe_score():
        def it_updates_uses(expect, locator):
            expect(locator.score(+1)) == True
            expect(locator.uses) == 1

        def it_tops_out_at_max_value(expect, locator):
            locator.score(+99, limit=42)
            expect(locator.score(+1, limit=42)) == False
            expect(locator.uses) == 42

        def it_bottoms_out_at_min_value(expect, locator):
            locator.score(-99, limit=-42)
            expect(locator.score(-1, limit=-42)) == False
            expect(locator.uses) == -42


def describe_action():
    def describe_str():
        def it_includes_the_verb_and_name(expect, action):
            expect(str(action)) == "fill_email"

    def describe_humanized():
        def it_describes_the_action(expect, action):
            expect(action.humanized) == "Filling email"

    def describe_sorted_locators():
        def it_orders_by_use(expect, action):
            action.locators = [
                Locator("id", "email", uses=-2),
                Locator("name", "email", uses=-3),
                Locator("value", "email", uses=-1),
            ]
            modes = [locator.mode for locator in action.sorted_locators]
            expect(modes) == ["value", "id", "name"]

        def it_tries_new_locators_first(expect, action):
            action.locators = [
                Locator("name", "email", uses=-1),
                Locator("id", "email", uses=0),
            ]
            expect(len(action.sorted_locators)) == 1
            expect(action.locator.mode) == "id"

        def it_only_includes_invalid_locators_when_no_other_options(expect, action):
            action.locators = [
                Locator("id", "email", uses=2),
                Locator("name", "email", uses=3),
                Locator("value", "email", uses=-1),
            ]
            modes = [locator.mode for locator in action.sorted_locators]
            expect(modes) == ["name", "id"]

    def describe_locator():
        def it_returns_placeholder_when_no_locators_defined(expect, action):
            action.locators = []
            expect(action.locator.value) == "placeholder"

    def describe_bool():
        def it_is_false_when_placeholder(expect, action):
            expect(bool(action)) == True
            action.name = ""
            expect(bool(action)) == False

    def describe_clean():
        def it_removes_unused_locators(expect, action):
            previous_count = len(action.locators)
            action.locators[0].uses = -1
            action.locators[1].uses = 2

            expect(action.clean("<page>")) == previous_count - 1
            expect(len(action.locators)) == 1

        def it_requires_one_locator_to_exceed_usage_threshold(expect, action):
            previous_count = len(action.locators)
            action.locators[0].uses = 1

            expect(action.clean("<page>")) == 0
            expect(len(action.locators)) == previous_count

        def it_can_be_forced(expect, action):
            previous_count = len(action.locators)
            action.locators[0].uses = 1

            expect(action.clean("<page>", force=True)) == previous_count - 1
            expect(len(action.locators)) == 1

        def it_keeps_used_locators(expect, action):
            action.locators = [Locator("mode", "value", uses=1)]

            expect(action.clean("<page>", force=True)) == 0
            expect(len(action.locators)) == 1


def describe_page():
    def describe_repr():
        def it_includes_the_url(expect, page):
            expect(repr(page)) == "Page.at('https://example.com/foo/bar')"

        def it_includes_the_variant_when_set(expect, page):
            page.variant = "qux"
            expect(
                repr(page)
            ) == "Page.at('https://example.com/foo/bar', variant='qux')"

    def describe_str():
        def it_returns_the_url(expect, page):
            expect(str(page)) == "https://example.com/foo/bar"

        def it_includes_the_variant_when_set(expect, page):
            page.variant = "qux"
            expect(str(page)) == "https://example.com/foo/bar (qux)"

    def describe_dir():
        def it_lists_valid_actions(expect, page, action):
            page.actions = []
            expect(dir(page)) == []
            page.actions.append(action)
            expect(dir(page)) == ["fill_email"]

    def describe_getattr():
        def it_returns_matching_action(expect, page, action):
            page.actions = [action]
            expect(getattr(page, "fill_email")) == action

        def it_adds_missing_actions(expect, page, monkeypatch):
            page.actions = []
            new_action = getattr(page, "fill_password")
            expect(new_action.verb) == "fill"
            expect(new_action.name) == "password"
            expect(len(new_action.locators)) > 1

        def it_rejects_invalid_actions(expect, page):
            with expect.raises(AttributeError):
                getattr(page, "mash_password")

        def it_rejects_missing_attributes(expect, page):
            with expect.raises(AttributeError):
                getattr(page, "foobar")

    def describe_contains():
        def it_matches_partial_html(expect, page, mockbrowser):
            expect(page).contains("world")

    def describe_clean():
        def it_removes_unused_locators(expect, page):
            page.locators.inclusions = [
                Locator("id", "foo", uses=0),
                Locator("id", "bar", uses=-1),
                Locator("id", "qux", uses=99),
            ]
            page.locators.exclusions = [
                Locator("id", "foo", uses=0),
                Locator("id", "bar", uses=-1),
                Locator("id", "qux", uses=99),
            ]

            expect(page.clean()) == 4
            expect(len(page.locators.inclusions)) == 1
            expect(len(page.locators.exclusions)) == 1

    def describe_properties():
        @pytest.mark.vcr()
        def it_computes_values_based_on_the_html(expect, page, mockbrowser):
            mockbrowser.html = requests.get("http://example.com", timeout=5).text
            expect(page.text) == (
                "Example Domain\n"
                "Example Domain\n"
                "This domain is for use in illustrative examples in documents. You may use this\n"
                "domain in literature without prior coordination or asking for permission.\n"
                "More information..."
            )
            expect(page.identity) == 19078



# pylint: disable=expression-not-assigned,unused-variable,redefined-outer-name,unused-argument

import log
import pytest

from ..types import URL, Fake


@pytest.fixture
def url():
    return URL("http://www.example.com/foo/bar")


def describe_url():
    def describe_init():
        def with_url(expect, url):
            expect(url.value) == "http://www.example.com/foo/bar"

        def with_url_and_trailing_slash(expect):
            url = URL("http://example.com/login/error/")
            expect(url.value) == "http://example.com/login/error"

        def with_path(expect):
            url = URL("example.com", "login")
            expect(url.value) == "https://example.com/login"

        def with_path_at_root(expect):
            url = URL("example.com", "@")
            expect(url.value) == "https://example.com"

        def with_path_and_extra_slashes(expect):
            url = URL("example.com", "/login/error/")
            expect(url.value) == "https://example.com/login/error"

    def describe_str():
        def it_returns_the_value(expect, url):
            expect(str(url)) == "http://www.example.com/foo/bar"

    def describe_eq():
        def it_compares_domain_and_path(expect, url):
            expect(url) == URL("https://www.example.com/foo/bar/")
            expect(url) != URL("http://example.com/foo/bar")

        def it_matches_patterns(expect):
            pattern = URL("http://example.com/p/{name}")
            expect(pattern) == URL("http://example.com/p/foobar")
            expect(pattern) != URL("http://example.com/")
            expect(pattern) != URL("http://example.com/p")
            expect(pattern) != URL("http://example.com/p/foo/bar")

        def it_does_not_match_root_placeholder(expect):
            pattern = URL("http://example.com/{value}")
            expect(pattern) != URL("http://example.com")

        def it_can_be_compared_to_str(expect, url):
            expect(url) == str(url)
            expect(url) == str(url) + "/"
            expect(url) != str(url) + "_extra"

    def describe_contains():
        def it_checks_url_contents(expect, url):
            expect(url).contains("foo/bar")
            expect(url).excludes("qux")

    def describe_path():
        def when_root(expect, url):
            url = URL("http://example.com")
            expect(url.path) == "@"

        def when_single(expect, url):
            url = URL("http://example.com/login")
            expect(url.path) == "login"

        def when_trailing_slash(expect, url):
            url = URL("http://example.com/login/error/")
            expect(url.path) == "login/error"

    def describe_fragment():
        def when_default(expect, url):
            expect(url.fragment) == ""

        def when_provided(expect, url):
            url = URL("http://example.com/signup/#step1")
            expect(url.fragment) == "step1"

        def when_containing_slashes(expect, url):
            url = URL("http://example.com/signup/#/step/2/")
            expect(url.fragment) == "step_2"

    def describe_detect_patterns():
        @pytest.mark.parametrize(
            ("before", "after"),
            [
                (
                    "https://example.com/items/42",
                    "https://example.com/items/{item}",
                ),
                (
                    "https://example.com/items/1/parts/2",
                    "https://example.com/items/{item}/parts/{part}",
                ),
                (
                    "https://example.com/42/",
                    "https://example.com/{id_1}",
                ),
                (
                    "https://example.com/items/123/456",
                    "https://example.com/items/{item}/{id_2}",
                ),
                (
                    "https://example.com/items/123/456/789",
                    "https://example.com/items/{item}/{id_2}/{id_3}",
                ),
                (
                    "https://example.com/items/20f5484b88ae49b08af03a389b4917dd",
                    "https://example.com/items/{item}",
                ),
                (
                    "https://en.wikipedia.org/wiki/Selenium_(software)",
                    "https://en.wikipedia.org/wiki/{wiki}",
                ),
            ],
        )
        def it_replaces_ids_with_placeholders(expect, before, after):
            url, updated = URL(before).detect_patterns()
            expect(url.value) == after


def describe_fake():
    @pytest.fixture
    def fake():
        return Fake()

    @pytest.mark.parametrize(
        "name",
        [
            "email_address",
            "zip_code",
        ],
    )
    def it_maps_aliases(expect, fake, name):
        value = getattr(fake, name)
        expect(value).isinstance(str)

    def it_handles_missing_attributes(expect, fake):
        with expect.raises(AttributeError):
            getattr(fake, "foobar")


def describe_person():
    @pytest.fixture
    def person():
        return Fake().person

    def it_includes_email_in_str(expect, person):
        expect(str(person)).contains("<")
        expect(str(person)).contains("@")

    def it_includes_name_in_email(expect, person):
        expect(person.email).icontains(person.last_name)

    def it_includes_state_name_and_abbr(expect, person):
        log.info(f"State: {person.state} ({person.state_abbr})")
        expect(len(person.state)) >= 4
        expect(len(person.state_abbr)) == 2

    @pytest.mark.parametrize(
        "name",
        [
            "address",
            "birthday",
            "cell_phone",
            "county",
            "email",
            "honorific",
            "nickname",
            "password",
            "phone",
            "prefix",
            "street_address",
            "username",
            "zip",
        ],
    )
    def it_maps_aliases(expect, person, name):
        value = getattr(person, name)
        expect(value).isinstance(str)

    def it_handles_missing_attributes(expect, person):
        with expect.raises(AttributeError):
            getattr(person, "foobar")



import inspect
from typing import Optional

import log

from . import models, prompts, types, utils
from .config import settings
from .models import Page, auto

__all__ = [
    "auto",
    "clean",
    "fake",
    "freeze",
    "log",
    "Page",
    "prompt",
    "visit",
]


fake = types.Fake()


def visit(
    url: str = "",
    *,
    framework: str = "",
    browser: str = "",
    headless: Optional[bool] = None,
    delay: float = 0.0,
) -> models.Page:
    if url:
        settings.url = url
    else:
        prompts.url_if_unset()

    if framework:
        settings.framework = framework.lower()
    else:
        prompts.framework_if_unset()

    if browser:
        settings.browser.name = browser.lower()
    else:
        prompts.browser_if_unset()

    if headless is not None:
        settings.browser.headless = headless

    utils.launch_browser(delay, silence_logging=True)
    utils.locate_models(caller=inspect.currentframe())

    return models.auto()


def prompt(*names):
    for name in names:
        prompts.secret_if_unset(name)
    return settings


def freeze():
    log.debug("Disabling interactive prompts and automatic locators")
    prompts.bullet = None
    settings.dev = False


def clean():
    utils.locate_models(caller=inspect.currentframe())
    for page in models.Page.objects.all():
        page.clean(force=True)



# pylint: disable=unused-variable,expression-not-assigned

from pomace import browser
from pomace.config import settings


def describe_launch():
    def it_requires_a_browser_to_be_set(expect):
        settings.browser.name = ""
        with expect.raises(SystemExit):
            browser.launch()

    def it_rejects_invalid_browsers(expect):
        settings.framework = "splinter"
        settings.browser.name = "foobar"
        with expect.raises(SystemExit):
            browser.launch()

    def it_forces_lowercase_browser_name(expect, mocker):
        settings.browser.name = "Firefox"
        mocker.patch.object(browser, "playwright")
        mocker.patch.object(browser.splinter, "Browser")
        browser.launch()
        expect(settings.browser.name) == "firefox"

    def it_handles_unspecified_browser(expect, mocker):
        settings.browser.name = "open"
        mocker.patch.object(browser, "playwright")
        mocker.patch.object(browser.splinter, "Browser")
        browser.launch()
        expect(settings.browser.name) == "firefox"
